# mypy: disable-error-code="override"
from typing import Optional

from klu.common.client import KluClientBase
from klu.data.constants import DATA_ENDPOINT
from klu.data.models import Data, DataSourceType
from klu.utils.dict_helpers import dict_no_empty


class DataClient(KluClientBase):
    def __init__(self, api_key: str):
        super().__init__(api_key, DATA_ENDPOINT, Data)

    async def create(
        self,
        input: str,
        output: str,
        action_guid: str,
        full_prompt_sent: Optional[str] = None,
        meta_data: Optional[dict] = None,
        session_guid: Optional[str] = None,
        model: Optional[str] = None,
        model_provider: Optional[str] = None,
        system_message: Optional[str] = None,
        latency: Optional[int] = None,
        num_output_tokens: Optional[int] = None,
        num_input_tokens: Optional[int] = None,
    ) -> Data:
        """
        Creates new data instance for an action_id provided.

        Args:
            issue (str): Data issue
            input (str): Input value of action execution
            output (str): The result of action execution
            rating (int): Data rating
            action_guid (str): Guid of an action data belongs to
            full_prompt_sent (str): String of a full prompt.
            session_guid (str): Guid of a session data belongs to
            meta_data (dict): Data meta_data
            correction (str): Data correction

        Returns:
            Created Data object
        """
        meta_data = meta_data or {}
        return await super().create(
            input=input,
            output=output,
            action=action_guid,
            session=session_guid,
            full_prompt_sent=full_prompt_sent,
            model=model,
            model_provider=model_provider,
            system_message=system_message,
            latency=latency,
            num_output_tokens=num_output_tokens,
            num_input_tokens=num_input_tokens,
            meta_data={
                **meta_data,
                "source": meta_data.pop("source", DataSourceType.SDK),
            },
        )

    async def get(self, guid: str) -> Data:
        """
        Retrieves data information based on the data ID.

        Args:
            guid (str): ID of a datum object to fetch.

        Returns:
            An object
        """
        return await super().get(guid)

    async def update(
        self,
        guid: str,
        meta_data: Optional[dict] = None,
        full_prompt_sent: Optional[str] = None,
        num_output_tokens: Optional[int] = None,
        num_input_tokens: Optional[int] = None,
        latency: Optional[int] = None,
        output: Optional[str] = None,
    ) -> Data:
        """
        Updated data information based on the data ID and provided payload. Currently, only supports `output` update.

        Args:

        Returns:
            Newly updated Data object
        """
        data = {
            "meta_data": meta_data,
            "full_prompt_sent": full_prompt_sent,
            "num_output_tokens": num_output_tokens,
            "num_input_tokens": num_input_tokens,
            "latency": latency,
            "output": output,
        }
        return await super().update(**{"guid": guid, **dict_no_empty(data)})

    async def delete(self, guid: str) -> Data:
        return await super().delete(guid)



from klu.common.errors import BaseKluError


class DataNotFoundError(BaseKluError):
    datum_id: int

    def __init__(self, datum_id):
        self.datum_id = datum_id
        self.message = f"Data with id {datum_id} was not found."
        super().__init__(self.message)



class ExperimentNotFoundError(Exception):
    experiment_id: str

    def __init__(self, experiment_id: str):
        self.experiment_id = experiment_id
        self.message = f"Experiment with id {experiment_id} was not found."
        super().__init__(self.message)


class InvalidExperimentPromptData(Exception):
    def __init__(self, response_message: str):
        self.message = (
            f"Failed to run the experiment due to the invalid request parameters. "
            f"Response message: {response_message}"
        )
        super().__init__(self.message)



"""
This module provides data models for the Data.
"""

from abc import ABC
from dataclasses import asdict, dataclass
from typing import Optional

from klu.common.models import BaseEngineModel, BasicStringEnum


@dataclass
class DataWithId(BaseEngineModel, ABC):
    """
    This class represents the base data with Id after it has been persisted into the Klu Database
    """

    guid: str

    def __repr__(self):
        return self.generate_repr()


@dataclass
class DataWithFeedbackUrl(BaseEngineModel, ABC):
    """
    This class represents the data specific to the model returned from the Action prompting
    """

    feedback_url: int

    def __repr__(self):
        return self.generate_repr()


@dataclass
class Feedback(BaseEngineModel):
    """
    This class represents the generic data information that is stored in the Klu database
    """

    guid: str
    type: str = ""
    value: str = ""
    data_guid: str = ""
    source: str = ""
    created_by: str = ""
    deleted: bool = False
    metadata: Optional[dict] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None

    def __repr__(self):
        return self.generate_repr()

    @classmethod
    def _from_engine_format(cls, data: dict) -> "Feedback":
        data["created_by"] = data.pop("createdById", None)
        data["data"] = data.pop("data_guid", None)
        data["metadata"] = data.pop("meta_data", None)
        return cls._create_instance(
            **{
                "updated_at": data.pop("updatedAt", None),
                "created_at": data.pop("createdAt", None),
            },
            **data,
        )

    def _to_engine_format(self) -> dict:
        base_dict = asdict(
            self, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}
        )

        base_dict.pop("updated_at", None)
        base_dict.pop("created_at", None)
        base_dict["createdById"] = base_dict.pop("created_by", None)
        base_dict["data"] = base_dict.pop("data_guid", None)
        base_dict["meta_data"] = base_dict.pop("metadata", None)
        return base_dict



# mypy: disable-error-code="override"
from typing import Any, List, Optional

import aiohttp
from aiohttp import ClientResponseError

from klu.common.client import KluClientBase
from klu.common.errors import (
    InvalidUpdateParamsError,
    UnknownKluAPIError,
    UnknownKluError,
)
from klu.common.models import TaskStatusEnum
from klu.context.constants import (
    CONTEXT_DOCUMENTS_ENDPOINT,
    CONTEXT_ENDPOINT,
    CONTEXT_PROMPT_ENDPOINT,
    CONTEXT_SEARCH_ENDPOINT,
    CONTEXT_STATUS_ENDPOINT,
    GENERATE_EMBEDDINGS,
    UPDATE_CONTEXT_DOCUMENT_ENDPOINT,
    UPLOAD_PRE_SIGNED_URL_ENDPOINT,
)
from klu.context.errors import ContextNotFoundError
from klu.context.models import (
    AddContextDocumentResponse,
    Context,
    ContextDocument,
    ContextPromptResponse,
    DeleteContextDocumentResponse,
    FileData,
    PreSignUrlPostData,
)
from klu.utils.dict_helpers import dict_no_empty
from klu.utils.file_upload import upload_to_pre_signed_url
from klu.workspace.errors import WorkspaceOrUserNotFoundError


class ContextClient(KluClientBase):
    def __init__(self, api_key: str):
        super().__init__(api_key, CONTEXT_ENDPOINT, Context)

    async def create(
        self,
        name: str,
        description: str,
        type: Optional[str] = None,
        loader: Optional[str] = None,
        filter: Optional[str] = None,
        meta_data: Optional[dict] = None,
        file_data: Optional[List[FileData]] = None,
        files: Optional[List[str]] = None,
    ) -> Context:
        """
        Creates a new context based on the provided data.

        Args:
            name (str): The name of the context.
            description (str): The description of the context.
            type (Optional[str]): The type of Context.
                Can be one of: [Simple, File, API, Loader]
            filter (Optional[str]): The filter to be used on Context
            loader (Optional[str]): The loader to be used on Context.
                Can be one of: [gmail, notion, intercom, github, redis, crawler, database, elastic, zendesk, file, custom]
                If missing, the API will use 'custom' as the default loader
            meta_data (Optional[dict]): The meta_data to be assigned to a created context
            file_data (Optional[FileData]): Metadata of the file to be uploaded.
                Can be omitted if only the context skeleton has to be created.
            files (Optional[List[str]]): List of s3 urls to be processed.

        Returns:
            The created Context object
        """
        file_urls = (
            [await self.upload_file(data) for data in file_data] if file_data else None
        )
        context = {
            "name": name,
            "type": type,
            "loader": loader,
            "filter": filter,
            "meta_data": meta_data,
            "description": description,
        }
        if file_urls:
            context["files"] = file_urls  # type: ignore
        elif files:
            context["files"] = files  # type: ignore

        return await super().create(**context)

    # type: ignore
    async def get(self, guid: str) -> Context:
        """
        Retrieves context information based on the id.

        Args:
            guid (str): id of a context object to fetch.

        Returns:
            Context object found by provided id
        """
        return await super().get(guid)

    # type: ignore
    async def update(
        self,
        context: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        meta_data: Optional[dict] = None,
    ) -> Context:
        """
        Update context data. At least one of the params has to be provided

        Args:
            guid (str): ID of a context to update.
            name: Optional[str]. New context name
            meta_data: Optional[dict]. New context meta_data. Can be used to update the files list on a context
            description: Optional[str]. New context description

        Returns:
            Updated app instance
        """

        if not name and not description and not meta_data:
            raise InvalidUpdateParamsError()

        return await super().update(
            **{
                "guid": context,
                **dict_no_empty(
                    {"name": name, "meta_data": meta_data, "description": description}
                ),
            }
        )

    # type: ignore
    async def add_files(self, context: str, files: List[str]) -> Context:
        """
        Add files to a context

        Args:
            context (str): guid of the context to update.
            files: List[str]. List of files to add to a context

        Returns:
            Updated context
        """

        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.post(
                    CONTEXT_ENDPOINT + f"{context}/add_files",
                    {"files": files},
                )
            except ClientResponseError as e:
                # TODO differentiate between missing context and missing user 404. Change the engine accordingly
                if e.status == 404:
                    raise WorkspaceOrUserNotFoundError()

                raise UnknownKluAPIError(e.status, e.message)
            except Exception as e:
                raise UnknownKluError(e)

            return Context._from_engine_format(response)

    # type: ignore
    async def delete(self, context: str) -> Context:
        """
        Delete existing context information defined by the id.

        Args:
            context (str): The guid of a context to delete.

        Returns:
            Deleted Context object
        """
        return await super().delete(guid=context)

    async def list(self) -> List[Context]:
        """
        Retrieves all actions for a user represented by the used API_KEY.
        Does not rely on internal paginator state, so `reset_pagination` method call can be skipped

        Returns (List[Action]): An array of all actions
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            response = await client.get(CONTEXT_ENDPOINT)

            return [Context._from_engine_format(context) for context in response]

    async def get_status(self, context: str) -> TaskStatusEnum:
        """
        Retrieves the status of an context creation task based on the provided context guid.

        Args:
            context (str): The guid of the context.

        Returns:
            string representing te context status
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.post(CONTEXT_STATUS_ENDPOINT.format(id=context))
            except ClientResponseError as e:
                if e.status == 404:
                    raise ContextNotFoundError(context)

                raise UnknownKluAPIError(e.status, e.message)

            return TaskStatusEnum.get(response.get("status"))  # type: ignore

    async def embed(
        self,
        context: str,
    ) -> dict:
        """
        Generates embeddings for the data in the context

        Args:
            context (str): Guid of context to process
        Returns:
            dict with a message about successful index creation
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.post(
                    GENERATE_EMBEDDINGS,
                    {
                        "guid": context,
                    },
                )
            except ClientResponseError as e:
                # TODO differentiate between missing context and missing user 404. Change the engine accordingly
                if e.status == 404:
                    raise WorkspaceOrUserNotFoundError()

                raise UnknownKluAPIError(e.status, e.message)
            except Exception as e:
                raise UnknownKluError(e)

            return {"message": response.get("msg")}

    async def upload_file(self, file_data: FileData) -> str:
        """
        Upload system file to Klu storage for later usage in context creation. Maximum supported file size is 50 MB.

        Args:
            file_data (FileData): Metadata of the file to be uploaded. For more details, see the FileData class docs.

        Returns:
            URL to the uploaded file. This URL can be used during the context update flow by passing it into the context meta_data.
        """
        async with aiohttp.ClientSession() as session:
            pre_signed_url_data = await self.get_pre_signed_url(file_data.file_name)
            await upload_to_pre_signed_url(
                session, pre_signed_url_data, file_data.file_path
            )

            return pre_signed_url_data.object_url

    async def get_pre_signed_url(self, file_name: str) -> PreSignUrlPostData:
        """
        Get pre-signed url to upload files to use for contexts creation. Maximum supported file size is 50 MB.
        This method should only be used if you don't want to use `upload_model_file` function to upload the file without
        the need to get into pre_signed_url upload flow.

        Args:
            file_name (str): The name of the file to be uploaded. Has to be unique among the files you uploaded before.
                Otherwise, the new file will override the previously uploaded one by the same file_name

        Returns:
            pre-signed url data including url, which is the pre-signed url that can be used to upload the file.
            Also includes 'fields' property that contains dict with data that
            has to be passed alongside the file during the upload
            And object_url property that contains the url that can be used to access the file location after the upload.
            This same object_url can be used during the context creation.
            For a usage example check out the `upload_index_file` function
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.post(
                    UPLOAD_PRE_SIGNED_URL_ENDPOINT,
                    {
                        "file_name": file_name,
                    },
                )
            except ClientResponseError as e:
                # TODO differentiate between missing context and missing user 404. Change the engine accordingly
                if e.status == 404:
                    raise WorkspaceOrUserNotFoundError()

                raise UnknownKluAPIError(e.status, e.message)

            return PreSignUrlPostData(**response)

    async def add_doc(
        self,
        context: str,
        content: str,
        filter: Optional[str] = "",
        meta_data: Optional[Any] = None,
    ) -> AddContextDocumentResponse:
        """
        Add a document to context

        Args:
            context (str): Guid of the context, which a document is created for
            text (str): Document text
            filter (str): Document filter
            meta_data: Optional[Any]. Document meta_data. Can be dict or array.

        Returns: response with context guid and status
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.post(
                    CONTEXT_DOCUMENTS_ENDPOINT.format(id=context),
                    {
                        "content": content,
                        "filter": filter,
                        "meta_data": meta_data,
                    },
                )
            except ClientResponseError as e:
                # TODO differentiate between missing context and missing user 404. Change the engine accordingly
                if e.status == 404:
                    raise ContextNotFoundError(context)

                raise UnknownKluAPIError(e.status, e.message)

            return AddContextDocumentResponse._create_instance(**response)

    async def list_docs(
        self,
        context: str,
    ) -> List[ContextDocument]:
        """
        List documents on a context

        Args:
            context(str): Guid of the context to fetch documents for

        Returns: response with the list of the documents on a context
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.get(
                    CONTEXT_DOCUMENTS_ENDPOINT.format(id=context),
                )
            except ClientResponseError as e:
                # TODO differentiate between missing context and missing user 404. Change the engine accordingly
                if e.status == 404:
                    raise ContextNotFoundError(context)

                raise UnknownKluAPIError(e.status, e.message)

            return [
                ContextDocument._from_engine_format(context_doc)
                for context_doc in response.get("data")
            ]

    async def delete_all_docs(
        self,
        context: str,
    ) -> DeleteContextDocumentResponse:
        """
        Delete all documents from a context. Executes in a task, so results will not be available immediately

        Args:
            context (str): Guid of the context to delete documents from

        Returns: response with a status message
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.delete(
                    CONTEXT_DOCUMENTS_ENDPOINT.format(id=context),
                )
            except ClientResponseError as e:
                # TODO differentiate between missing context and missing user 404. Change the engine accordingly
                if e.status == 404:
                    raise ContextNotFoundError(context)

                raise UnknownKluAPIError(e.status, e.message)

            return DeleteContextDocumentResponse(**response)

    async def update_doc(
        self,
        doc: str,
        context: str,
        content: str,
        embedding: List[float],
        filter: Optional[str] = None,
        meta_data: Optional[Any] = None,
    ) -> ContextDocument:
        """
        Update a single document on a context

        Args:
            doc (str): Guid of the doc to update
            context (str): Guid of the context the doc belongs to
            content (str): New content of a doc
            embedding (list[float]): New embedding list of a doc
            filter (Optional[str]): New filter of a doc
            meta_data (Optional[Any]): new meta_data of a doc

        Returns: updated context object
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.put(
                    UPDATE_CONTEXT_DOCUMENT_ENDPOINT.format(id=context, doc_id=doc),
                    {
                        "content": content,
                        "embedding": embedding,
                        **dict_no_empty(
                            {
                                "filter": filter,
                                "meta_data": meta_data,
                            }
                        ),
                    },
                )
            except ClientResponseError as e:
                # TODO differentiate between missing context and missing user 404. Change the engine accordingly
                if e.status == 404:
                    raise ContextNotFoundError(context)

                raise UnknownKluAPIError(e.status, e.message)

            return ContextDocument._from_engine_format(response)

    async def search(
        self,
        context: str,
        prompt: str,
        similarity_score: Optional[float] = 0,
        number_of_results: Optional[int] = 10,
        filter: Optional[str] = None,
        metadata: Optional[Any] = None,
    ) -> List[ContextDocument]:
        """
        Search Context

        Args:
            context (str): Guid of the context to search
            prompt (str): Prompt to search
            similarity_score (Optional[float]): Minimum similarity score to return
            number_of_results (Optional[int]): Number of results to return

        Returns: List of relevant documents
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.post(
                    CONTEXT_SEARCH_ENDPOINT.format(id=context),
                    {
                        "prompt": prompt,
                        "similarity_score": similarity_score,
                        "number_of_results": number_of_results,
                        "filter": filter,
                        "metadata_filter": metadata,
                    },
                )
            except ClientResponseError as e:
                # TODO differentiate between missing context and missing user 404. Change the engine accordingly
                if e.status == 404:
                    raise ContextNotFoundError(context)

                raise UnknownKluAPIError(e.status, e.message)

            return [
                ContextDocument._from_engine_format(context_doc)
                for context_doc in response
            ]

    async def prompt(
        self,
        context: str,
        prompt: str,
        response_mode: Optional[str] = "search",
        response_length: Optional[int] = 1024,
        number_of_results: Optional[int] = 5,
    ) -> ContextPromptResponse:
        """
        Search Context

        Args:
            context (str): Guid of the context to search
            prompt (str): Prompt to search
            similarity_score (Optional[float]): Minimum similarity score to return
            number_of_results (Optional[int]): Number of results to return

        Returns: List of relevant documents
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.post(
                    CONTEXT_PROMPT_ENDPOINT,
                    {
                        "context": context,
                        "prompt": prompt,
                        "responseMode": response_mode,
                        "responseLength": response_length,
                        "similarityTopK": number_of_results,
                    },
                )
            except ClientResponseError as e:
                # TODO differentiate between missing context and missing user 404. Change the engine accordingly
                if e.status == 404:
                    raise ContextNotFoundError(context)

                raise UnknownKluAPIError(e.status, e.message)

            return ContextPromptResponse(**response)



# mypy: disable-error-code="override"
from typing import NoReturn, Optional

from klu.common.client import KluClientBase
from klu.common.errors import NotSupportedError
from klu.session.constants import SESSION_ENDPOINT
from klu.session.models import Session


class SessionClient(KluClientBase):
    def __init__(self, api_key: str):
        super().__init__(api_key, SESSION_ENDPOINT, Session)

    async def create(
        self, action: str, name: Optional[str] = None, extUserId: Optional[str] = None
    ) -> Session:
        """
        Creates a session based on the data provided.

        Args:
            action_guid (str): Guid of the action object this session is attached to.
            name Optional[str]: Name of the session. If not provided will be left blank.

        Returns:
            A newly created Session object.
        """
        return await super().create(action=action, name=name, extUserId=extUserId)

    async def get(self, guid: str) -> Session:
        """
        Retrieves session information based on the unique Session guid, created during the Session creation.

        Args:
            guid (str): GUID of a session to fetch. The one that was used during the session creation

        Returns:
            Session object
        """
        return await super().get(guid)

    async def update(self) -> NoReturn:
        raise NotSupportedError()

    async def delete(self, guid: str) -> Session:
        """
        Delete session based on the id.

        Args:
            guid (str): Unique Guid of a session to delete.

        Returns:
            Deleted session object
        """
        return await super().delete(guid)



"""
This module provides data models for the Action.
"""
from dataclasses import asdict, dataclass
from typing import Any, AsyncIterator, Dict, List, Optional

from klu.api.sse_client import SSEClient
from klu.common.models import (
    BaseDataClass,
    BaseEngineModel,
    PromptInput,
    TaskStatusEnum,
)


@dataclass
class ModelConfig(BaseDataClass):
    """
    This class represents the ModelConfig data model returned from the Klu engine
    """

    top_p: Optional[float] = 1
    temperature: Optional[float] = 0.5
    timeout: Optional[int] = 60
    num_retries: Optional[int] = 2
    stop_sequence: Optional[List[str]] = None
    presence_penalty: Optional[float] = 0
    frequency_penalty: Optional[float] = 0
    max_response_length: Optional[int] = 1000


@dataclass
class Action(BaseEngineModel):
    """
    This class represents the Action data model returned from the Klu engine
    """

    guid: str
    name: str
    app: str
    model: str
    action_type: str
    prompt: PromptInput
    system_message: Optional[str]
    description: Optional[str]
    model_config: Optional[ModelConfig]
    updated_at: Optional[str] = None
    created_at: Optional[str] = None
    meta_data: Optional[Dict[str, Any]] = None

    def __repr__(self):
        return self.generate_repr()

    @classmethod
    def _from_engine_format(cls, data: dict) -> "Action":
        model_config = data.pop("model_config", {})
        if model_config is None:
            model_config = {}
        mc_config = {
            "top_p": model_config.get("topP", None),
            "temperature": model_config.get("temperature", None),
            "timeout": model_config.get("timeout", None),
            "num_retries": model_config.get("numRetries", None),
            "stop_sequence": model_config.get("stopSequence", []),
            "presence_penalty": model_config.get("presencePenalty", None),
            "frequency_penalty": model_config.get("frequencyPenalty", None),
            "max_response_length": model_config.get("maxResponseLength", None),
        }
        data["model_config"] = ModelConfig(**mc_config)
        obj = cls._create_instance(
            **{
                "updated_at": data.pop("updatedAt", None),
                "created_at": data.pop("createdAt", None),
            },
            **data,
        )
        return obj

    def _to_engine_format(self) -> dict:
        base_dict = asdict(self)
        model_config = base_dict.pop("model_config", None)
        return {
            "meta_data": base_dict.pop("meta_data", None),
            "updatedAt": base_dict.pop("updated_at", None),
            "createdAt": base_dict.pop("created_at", None),
            "model_config": {
                "topP": model_config.top_p,
                "temperature": model_config.temperature,
                "timeout": model_config.timeout,
                "numRetries": model_config.num_retries,
                "stopSequence": model_config.stop_sequence,
                "presencePenalty": model_config.presence_penalty,
                "frequencyPenalty": model_config.frequency_penalty,
                "maxResponseLength": model_config.max_response_length,
            },
            **base_dict,
        }


@dataclass
class PromptResponse(BaseDataClass):
    """
    This class represents the Response data model returned from the Klu engine in response to action prompting.
    """

    msg: str
    streaming: bool
    data_guid: Optional[str] = None
    result_url: Optional[str] = None
    feedback_url: Optional[str] = None
    streaming_url: Optional[str] = None

    def __repr__(self):
        return self.generate_repr()


@dataclass
class SyncPromptResponse(BaseDataClass):
    """
    This class represents the Response data model returned
    from the Klu engine in response to non-async, non-streaming action prompting.
    """

    msg: str
    data_guid: Optional[str] = None
    feedback_url: Optional[str] = None

    def __repr__(self):
        return self.generate_repr()


@dataclass
class StreamingPromptResponse(BaseDataClass):
    """
    This class represents the Response data model returned from the Klu engine in response to streaming action prompting.
    """

    msg: str
    streaming_url: Optional[str]
    data_guid: Optional[str] = None

    feedback_url: Optional[str] = None
    sse_client: Optional[SSEClient] = None

    @classmethod
    def _create_instance(cls, **kwargs):
        instance = cls.__new__(cls)
        instance._init_with_base_class(**kwargs)
        return instance

    def _init_with_base_class(self, **kwargs):
        super()._init_with_base_class(**kwargs)
        if self.streaming_url:
            self.sse_client = SSEClient(self.streaming_url)

    @property
    def tokens(self) -> Optional[AsyncIterator[str]]:
        return self.sse_client.get_streaming_data() if self.sse_client else None

    def __repr__(self):
        return self.generate_repr()


@dataclass
class AsyncPromptResponse(BaseDataClass):
    """
    This class represents the Response data model returned from the Klu engine in response to async action prompting.
    Contains result_url  - the url that gives access to the result when the prompt is completed or a message about prompting in progress.
    """

    msg: str
    data_guid: Optional[str] = None
    result_url: Optional[str] = None
    feedback_url: Optional[str] = None

    def __repr__(self):
        return self.generate_repr()


@dataclass
class AsyncPromptResultResponse(BaseDataClass):
    """
    This class represents the Response data model returned from the Klu engine in response to triggering result_url
    returned from the async prompting endpoint.
    msg will contain the result once the prompting task is completed.
    """

    msg: str
    status: TaskStatusEnum

    def __repr__(self):
        return self.generate_repr()



# mypy: disable-error-code="override"
from typing import List, Optional

import aiohttp
from aiohttp import ClientResponseError

from klu.action.constants import (
    ACTION_CONTEXT_ENDPOINT,
    ACTION_DATA_ENDPOINT,
    ACTION_ENDPOINT,
    ACTION_SKILL_ENDPOINT,
    CREATE_ACTION_ENDPOINT,
    DEFAULT_ACTIONS_PAGE_SIZE,
    DEPLOY_ACTION_ENDPOINT,
    PLAYGROUND_PROMPT_ENDPOINT,
)
from klu.action.errors import ActionNotFoundError, InvalidActionPromptData
from klu.action.models import (
    Action,
    AsyncPromptResponse,
    AsyncPromptResultResponse,
    ModelConfig,
    PromptResponse,
    StreamingPromptResponse,
    SyncPromptResponse,
)
from klu.common.client import KluClientBase
from klu.common.errors import (
    InvalidUpdateParamsError,
    UnknownKluAPIError,
    UnknownKluError,
)
from klu.common.models import PromptInput
from klu.context.models import Context
from klu.data.models import Data
from klu.skill.models import Skill
from klu.utils.dict_helpers import dict_no_empty
from klu.utils.paginator import Paginator
from klu.workspace.errors import WorkspaceOrUserNotFoundError


class ActionsClient(KluClientBase):
    def __init__(self, api_key: str):
        super().__init__(api_key, ACTION_ENDPOINT, Action)
        self._paginator = Paginator(ACTION_ENDPOINT)

    # type: ignore
    async def create(
        self,
        name: str,
        prompt: str,
        description: str,
        app_guid: str,
        model_guid: str,
        action_type: Optional[str] = "prompt",
        system_message: Optional[str] = None,
        model_config: Optional[ModelConfig] = None,
    ) -> Action:
        """
        Creates new action instance

        Args:
            name (str): Action name
            prompt (str): Action prompt
            model_guid (int): Guid of a model used for action
            app_guid (str): GUID of the app for an action to be attached to
            action_type (str): The type of the action. Can be one of ['prompt', 'chat', 'workflow', 'worker', 'type']
            description (str): The description of the action
            model_config (ModelConfig): Optional action model configuration dict.

        Returns:
            Newly created Action object
        """
        return await super().create(
            name=name,
            prompt=prompt,
            app=app_guid,
            model=model_guid,
            action_type=action_type,
            description=description,
            model_config=model_config if model_config else None,
            system_message=system_message,
            url=CREATE_ACTION_ENDPOINT,
        )

    # type: ignore
    async def get(self, guid: str) -> Action:
        """
        Get an action defined by the id

        Args:
            guid (str): The id of an action to retrieve

        Returns:
            Retrieved Action object.
        """
        return await super().get(guid)

    # type: ignore
    async def update(
        self,
        action: str,
        name: Optional[str] = None,
        prompt: Optional[str] = None,
        description: Optional[str] = None,
        model_config: Optional[str] = None,
        system_message: Optional[str] = None,
    ) -> Action:
        """
        Update action instance with provided data. At least one of parameters should be present.

        Args:
            guid (str): The GUID of the action to update.
            name: Optional[str]. New action name.
            prompt: Optional[str]. New action type.
            description: Optional[str]. New action description.
            model_config: Optional[dict]. New action model_config.

        Returns:
            Action with updated data
        """
        if not name and not prompt and not description and not model_config:
            raise InvalidUpdateParamsError()

        return await super().update(
            **{
                "guid": action,
                **dict_no_empty(
                    {
                        "name": name,
                        "prompt": prompt,
                        "description": description,
                        "model_config": model_config,
                        "system_message": system_message,
                    }
                ),
            }
        )

    async def deploy(
        self,
        action: str,
    ) -> bool:
        """
        Deploy action

        Args:
            guid (str): The GUID of the action to update.

        Returns:
            Boolean value indicating whether the action was successfully deployed
        """

        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)

            try:
                response = await client.get(
                    DEPLOY_ACTION_ENDPOINT.format(id=action),
                )
                return response
            except ClientResponseError as e:
                if e.status == 404:
                    raise WorkspaceOrUserNotFoundError()
                if e.status == 400:
                    raise InvalidActionPromptData(e.message)

                raise UnknownKluAPIError(e.status, e.message)
            except Exception as e:
                raise UnknownKluError(e)

    async def delete(self, action: str) -> Action:
        """
        Delete an action defined by the id

        Args:
            guid (str): The id of an action to delete

        Returns:
            Deleted Action object.
        """
        return await super().delete(action)

    async def prompt(
        self,
        action: str,
        input: PromptInput,
        filter: Optional[str] = None,
        session: Optional[str] = None,
        metadata_filter: Optional[dict] = None,
        cache: Optional[bool] = False,
    ) -> SyncPromptResponse:
        """
        Run a prompt with an agent, optionally using streaming.

        Args:
            action (str): The GUID of the agent to run the prompt with.
            input (PromptInput): The prompt to run with the agent.
            filter (Optional[str]): The filter to use when running the prompt.
            session (Optional[str]): The GUID of the session to run the prompt with.

        Returns:
            An object result of running the prompt with the message and a feedback_url for providing feedback.
        """
        response = await self._run_prompt(
            action=action,
            input=input,
            filter=filter,
            streaming=False,
            async_mode=False,
            session=session,
            metadata_filter=metadata_filter,
            cache=cache,
        )
        return SyncPromptResponse._create_instance(**response.__dict__)

    async def async_prompt(
        self,
        action: str,
        input: PromptInput,
        filter: Optional[str] = None,
        session: Optional[str] = None,
    ) -> AsyncPromptResponse:
        """
        Run a prompt with an agent, optionally using streaming.

        Args:
            action (str): The GUID of the agent to run the prompt with.
            input (PromptInput): The prompt to run with the agent.
            filter (Optional[str]): The filter to use when running the prompt.
            session (Optional[str]): The GUID of the session to run the prompt with.

        Returns:
            An object result of running the prompt with the message and a feedback_url for providing feedback.
            Also contains result_url - the url that gives access to the result when the prompt is completed or a message about prompting in progress.
        """
        response = await self._run_prompt(
            action=action,
            input=input,
            filter=filter,
            streaming=False,
            async_mode=True,
            session=session,
        )
        return AsyncPromptResponse._create_instance(**response.__dict__)

    async def get_async_prompt_result(
        self,
        result_url: str,
    ) -> AsyncPromptResultResponse:
        """
        Get a result of async prompting

        Args:
            result_url (str): The url you received in response to calling async_prompt function

        Returns:
            An object with the message, which contains either prompt result or a message about prompt still being in progress.
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)

            response = await client.get(result_url, api_url="")
            return AsyncPromptResultResponse._create_instance(**response)

    async def stream(
        self,
        action: str,
        input: PromptInput,
        filter: Optional[str] = None,
        session: Optional[str] = None,
    ) -> StreamingPromptResponse:
        """
        Run a prompt with an agent, optionally using streaming.

        Args:
            action (str): The GUID of the agent to run the prompt with.
            input (PromptInput): The prompt to run with the agent.
            filter (Optional[str]): The filter to use when running the prompt.
            session (Optional[str]): The GUID of the session to run the prompt with.

        Returns:
            An object result of running the prompt with the message, feedback_url for providing feedback.
            The response will also contain data stream, which can be used to consume the prompt response message
        """
        prompt_response = await self._run_prompt(
            action,
            input=input,
            filter=filter,
            streaming=True,
            async_mode=False,
            session=session,
        )

        return StreamingPromptResponse._create_instance(**prompt_response.__dict__)

    async def _run_prompt(
        self,
        action: str,
        input: PromptInput,
        filter: Optional[str] = None,
        streaming: Optional[bool] = False,
        async_mode: Optional[bool] = False,
        session: Optional[str] = None,
        metadata_filter: Optional[dict] = None,
        cache: Optional[bool] = False,
    ) -> PromptResponse:
        async with aiohttp.ClientSession() as _session:
            client = self._get_api_client(_session)

            try:
                response = await client.post(
                    ACTION_ENDPOINT,
                    {
                        "input": input,
                        "filter": filter,
                        "action": action,
                        "streaming": streaming,
                        "async_mode": async_mode,
                        "session": session,
                        "metadata_filter": metadata_filter,
                        "cache": cache,
                    },
                )
                return PromptResponse(**response)
            except ClientResponseError as e:
                if e.status == 404:
                    raise ActionNotFoundError(action)

                raise UnknownKluAPIError(e.status, e.message)
            except Exception as e:
                raise UnknownKluError(e)

    async def playground(
        self,
        prompt: str,
        model_id: int,
        values: Optional[dict] = None,
        tool_ids: Optional[list] = None,
        index_ids: Optional[list] = None,
        model_config: Optional[dict] = None,
    ) -> StreamingPromptResponse:
        """
        Run a prompt with an agent. Always uses streaming

        Args:
            prompt (str): The prompt to run.
            model_id (int): The ID of the model to use. Can be retrieved by querying the model by guid
            tool_ids (list): Optional list of tool IDs to use. Defaults to an empty array
            index_ids (list): Optional list of index IDs to use. Defaults to an empty array
            values (Optional[dict]): The values to be interpolated into the prompt template, or appended to the prompt template if it doesn't include variables
            model_config (Optional[dict]): Configuration of the model

        Returns:
            An object result of running the prompt with the message and a feedback_url for providing feedback.
        """
        tool_ids = tool_ids or []
        index_ids = index_ids or []

        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)

            try:
                response = await client.post(
                    PLAYGROUND_PROMPT_ENDPOINT,
                    {
                        "prompt": prompt,
                        "values": values,
                        "toolIds": tool_ids,
                        "modelId": model_id,
                        "indexIds": index_ids,
                        "modelConfig": model_config,
                    },
                )
                return StreamingPromptResponse._create_instance(**response)
            except ClientResponseError as e:
                if e.status == 404:
                    raise WorkspaceOrUserNotFoundError()
                if e.status == 400:
                    raise InvalidActionPromptData(e.message)

                raise UnknownKluAPIError(e.status, e.message)
            except Exception as e:
                raise UnknownKluError(e)

    async def get_data(self, guid: str) -> List[Data]:
        """
        Retrieves data information for an action.

        Args:
            guid (str): Guid of an action to fetch data for.

        Returns:
            An array of actions found by provided app id.
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)

            try:
                response = await client.get(ACTION_DATA_ENDPOINT.format(id=guid))
                return [Data._from_engine_format(data) for data in response]
            except ClientResponseError as e:
                if e.status == 404:
                    raise ActionNotFoundError(guid)

                raise UnknownKluAPIError(e.status, e.message)
            except Exception as e:
                raise UnknownKluError(e)

    async def list(self) -> List[Action]:
        """
        Retrieves all actions for a user represented by the used API_KEY.
        Does not rely on internal paginator state, so `reset_pagination` method call can be skipped

        Returns (List[Action]): An array of all actions
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            response = await self._paginator.fetch_all(client)

            return [Action._from_engine_format(action) for action in response]

    async def fetch_single_page(
        self, page_number, limit: int = DEFAULT_ACTIONS_PAGE_SIZE
    ) -> List[Action]:
        """
        Retrieves a single page of actions.
        Can be used to fetch a specific page of actions provided a certain per_page config.
        Does not rely on internal paginator state, so `reset_pagination` method call can be skipped

        Args:
            page_number (int): Number of the page to fetch
            limit (int): Number of instances to fetch per page. Defaults to 50

        Returns:
            An array of actions fetched for a queried page. Empty if queried page does not exist
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            response = await self._paginator.fetch_single_page(
                client, page_number, limit=limit
            )

            return [Action._from_engine_format(action) for action in response]

    async def fetch_next_page(
        self, limit: int = DEFAULT_ACTIONS_PAGE_SIZE, offset: Optional[int] = None
    ) -> List[Action]:
        """
        Retrieves the next page of actions. Can be used to fetch a flexible number of pages starting.
        The place to start from can be controlled by the offset parameter.
        After using this method, we suggest to call `reset_pagination` method to reset the page cursor.

        Args:
            limit (int): Number of instances to fetch per page. Defaults to 50
            offset (int): The number of instances to skip. Can be used to query the pages of actions skipping certain number of instances.

        Returns:
            An array of actions fetched for a queried page. Empty if the end was reached at the previous step.
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            response = await self._paginator.fetch_next_page(
                client, limit=limit, offset=offset
            )

            return [Action._from_engine_format(action) for action in response]

    async def reset_pagination(self):
        self._paginator = Paginator(ACTION_ENDPOINT)

    async def get_skills(self, guid: str) -> List[Skill]:
        """
        Retrieves all skills attached to an action.
        Args:
            guid (str): Action guid


        Returns (List[Skill]): An array of all skills
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)

            try:
                response = await client.get(ACTION_SKILL_ENDPOINT.format(id=guid))
                return [Skill._from_engine_format(skill) for skill in response]
            except ClientResponseError as e:
                if e.status == 404:
                    raise ActionNotFoundError(guid)

                raise UnknownKluAPIError(e.status, e.message)
            except Exception as e:
                raise UnknownKluError(e)

    async def update_skills(self, guid: str, skills: List[str]) -> bool:
        """
        Updates skills attached to an action
        Args:
            guid (str): Action guid
            skills (List[str]): An array of skill guids to attach to an action

        Returns (List[Skill]): An array of all skills
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)

            try:
                response = await client.put(
                    ACTION_SKILL_ENDPOINT.format(id=guid), {"skills": skills}
                )
                return response
            except ClientResponseError as e:
                if e.status == 404:
                    raise ActionNotFoundError(guid)

                raise UnknownKluAPIError(e.status, e.message)
            except Exception as e:
                raise UnknownKluError(e)

    async def get_context(self, guid) -> List[Context]:
        """
        Retrieves all context attached to an action
        Args:
            guid (str): Action guid

        Returns (List[Context]): An array of all context
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)

            try:
                response = await client.get(ACTION_CONTEXT_ENDPOINT.format(id=guid))
                return [Context._from_engine_format(context) for context in response]
            except ClientResponseError as e:
                if e.status == 404:
                    raise ActionNotFoundError(guid)

                raise UnknownKluAPIError(e.status, e.message)
            except Exception as e:
                raise UnknownKluError(e)

    async def upddate_context(self, guid: str, context: List[str]) -> bool:
        """
        Retrieves all context attached to an action
        Args:
            guid (str): Action guid
            context (List[str]): An array of context guids to attach to an action

        Returns (List[Context]): An array of all context
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)

            try:
                response = await client.put(
                    ACTION_CONTEXT_ENDPOINT.format(id=guid), {"context": context}
                )
                return response
            except ClientResponseError as e:
                if e.status == 404:
                    raise ActionNotFoundError(guid)

                raise UnknownKluAPIError(e.status, e.message)
            except Exception as e:
                raise UnknownKluError(e)



# mypy: disable-error-code="override"
from typing import List, Optional

import aiohttp
from aiohttp import ClientResponseError

from klu.action.models import (
    AsyncPromptResponse,
    AsyncPromptResultResponse,
    PromptResponse,
    StreamingPromptResponse,
    SyncPromptResponse,
)
from klu.common.client import KluClientBase
from klu.common.errors import (
    InvalidUpdateParamsError,
    UnknownKluAPIError,
    UnknownKluError,
)
from klu.common.models import PromptInput
from klu.experiment.constants import EXPERIMENT_ENDPOINT, PROMPT_EXPERIMENT_ENDPOINT
from klu.experiment.errors import ExperimentNotFoundError
from klu.experiment.models import Experiment
from klu.utils.dict_helpers import dict_no_empty


class ExperimentClient(KluClientBase):
    def __init__(self, api_key: str):
        super().__init__(api_key, EXPERIMENT_ENDPOINT, Experiment)

    async def create(
        self,
        name: str,
        app_guid: str,
        action_primary_guid: str,
        action_secondary_guid: str,
    ) -> Experiment:
        """
        Creates an experiment based on the data provided.

        Args:
            name (str): Name of the experiment
            app_guid (str): Guid of the app object this experiment is attached to.
            action_primary_guid (str): Guid of the primary action object this experiment is attached to.
            action_secondary_guid (str): Guid of the secondary action object this experiment is attached to.

        Returns:
            A newly created Experiment object.
        """
        return await super().create(
            name=name,
            app_guid=app_guid,
            action_primary_guid=action_primary_guid,
            action_secondary_guid=action_secondary_guid,
        )

    async def get(self, guid: str) -> Experiment:
        """
        Retrieves experiment information based on the unique Experiment guid, created during the Experiment creation.

        Args:
            guid (str): GUID of an experiment to fetch. The one that was used during the experiment creation

        Returns:
            Experiment object
        """
        return await super().get(guid)

    async def update(
        self,
        guid: str,
        name: Optional[str] = None,
        action_primary_guid: Optional[str] = None,
        action_secondary_guid: Optional[str] = None,
    ) -> Experiment:
        """
        Update experiment data. At least one of the params has to be provided

        Args:
            guid (str): GUID of an experiment to update.
            name: Optional[str]. New experiment name
            action_primary_guid: Optional[str]. New experiment primary action guid
            action_secondary_guid: Optional[str]. New experiment secondary action guid

        Returns:
            Updated experiment instance
        """

        if not name and not action_primary_guid and not action_secondary_guid:
            raise InvalidUpdateParamsError()

        return await super().update(
            **{
                "guid": guid,
                **dict_no_empty(
                    {
                        "name": name,
                        "action_primary_guid": action_primary_guid,
                        "action_secondary_guid": action_secondary_guid,
                    }
                ),
            }
        )

    async def delete(self, guid: str) -> Experiment:
        """
        Delete experiment based on the id.

        Args:
            guid (str): Unique Guid of an experiment to delete.

        Returns:
            Deleted experiment object
        """
        return await super().delete(guid)

    async def prompt(
        self,
        experiment_guid: str,
        input: PromptInput,
        filter: Optional[str] = None,
        force_action: Optional[str] = None,
        session_guid: Optional[str] = None,
        indices: Optional[List[int]] = None,
        metadata_filter: Optional[dict] = None,
    ) -> SyncPromptResponse:
        """
        Run a prompt for an experiment in a sync non-streaming mode

        Args:
            experiment_guid (str): The GUID of the experiment to run the prompt with.
            input (PromptInput): The experiment prompt input. Can be string or a key-value dict.
            filter (Optional[str]): The filter to use when running the prompt.
            force_action (Optional[str]): The action guid to enforce when running experiment prompt.
            session_guid (Optional[str]): The GUID of the session to run the prompt with.
            indices (Optional[List[int]]): An array of indices to be used when running the prompt.
            metadata_filter (Optional[dict]): The filter over the metadata to be applied during prompting

        Returns:
            An object result of running the prompt with the message and a feedback_url for providing feedback.
        """
        response = await self._run_prompt(
            experiment_guid,
            input,
            streaming=False,
            async_mode=False,
            filter=filter,
            indices=indices,
            force_action=force_action,
            session_guid=session_guid,
            metadata_filter=metadata_filter,
        )
        return SyncPromptResponse._create_instance(**response.__dict__)

    async def async_prompt(
        self,
        experiment_guid: str,
        input: PromptInput,
        filter: Optional[str] = None,
        force_action: Optional[str] = None,
        session_guid: Optional[str] = None,
        indices: Optional[List[int]] = None,
        metadata_filter: Optional[dict] = None,
    ) -> AsyncPromptResponse:
        """
        Run a prompt for an experiment in an async mode.

        Args:
            experiment_guid (str): The GUID of the experiment to run the prompt with.
            input (PromptInput): The experiment prompt input. Can be string or a key-value dict.
            filter (Optional[str]): The filter to use when running the prompt.
            force_action (Optional[str]): The action guid to enforce when running experiment prompt.
            session_guid (Optional[str]): The GUID of the session to run the prompt with.
            indices (Optional[List[int]]): An array of indices to be used when running the prompt.
            metadata_filter (Optional[dict]): The filter over the metadata to be applied during prompting

        Returns:
            An object result of running the prompt with the message about successful start of prompting and a feedback_url for providing feedback.
            Also contains result_url - the url that gives access to the result when the prompt is completed or a message about prompting in progress.
        """
        response = await self._run_prompt(
            experiment_guid,
            input,
            streaming=False,
            async_mode=True,
            filter=filter,
            indices=indices,
            force_action=force_action,
            session_guid=session_guid,
            metadata_filter=metadata_filter,
        )
        return AsyncPromptResponse._create_instance(**response.__dict__)

    async def get_async_prompt_result(
        self,
        result_url: str,
    ) -> AsyncPromptResultResponse:
        """
        Get a result of async prompting

        Args:
            result_url (str): The url you received in response to calling async_prompt function

        Returns:
            An object with the message, which contains either prompt result or a message about prompt still being in progress.
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)

            response = await client.get(result_url, api_url="")
            return AsyncPromptResultResponse._create_instance(**response)

    async def stream(
        self,
        experiment_guid: str,
        input: PromptInput,
        filter: Optional[str] = None,
        force_action: Optional[str] = None,
        session_guid: Optional[str] = None,
        indices: Optional[List[int]] = None,
        metadata_filter: Optional[dict] = None,
    ) -> StreamingPromptResponse:
        """
        Run a prompt for an experiment in a streaming mode.

        Args:
            experiment_guid (str): The GUID of the experiment to run the prompt with.
            input (PromptInput): The experiment prompt input. Can be string or a key-value dict.
            filter (Optional[str]): The filter to use when running the prompt.
            force_action (Optional[str]): The action guid to enforce when running experiment prompt.
            session_guid (Optional[str]): The GUID of the session to run the prompt with.
            indices (Optional[List[int]]): An array of indices to be used when running the prompt.
            metadata_filter (Optional[dict]): The filter over the metadata to be applied during prompting

        Returns:
            An object result of running the prompt with the message, feedback_url for providing feedback.
            The response will also contain data stream, which can be used to consume the prompt response message
        """
        prompt_response = await self._run_prompt(
            experiment_guid,
            input,
            streaming=True,
            async_mode=False,
            filter=filter,
            indices=indices,
            force_action=force_action,
            session_guid=session_guid,
            metadata_filter=metadata_filter,
        )

        return StreamingPromptResponse._create_instance(**prompt_response.__dict__)

    async def _run_prompt(
        self,
        experiment_guid: str,
        input: PromptInput,
        filter: Optional[str] = None,
        force_action: Optional[str] = None,
        streaming: Optional[bool] = False,
        async_mode: Optional[bool] = False,
        session_guid: Optional[str] = None,
        indices: Optional[List[int]] = None,
        metadata_filter: Optional[dict] = None,
    ) -> PromptResponse:
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)

            try:
                response = await client.post(
                    PROMPT_EXPERIMENT_ENDPOINT.format(id=experiment_guid),
                    {
                        "input": input,
                        "filter": filter,
                        "indices": indices,
                        "streaming": streaming,
                        "session": session_guid,
                        "async_mode": async_mode,
                        "force_action": force_action,
                        "metadata_filter": metadata_filter,
                    },
                )
                return PromptResponse(**response)
            except ClientResponseError as e:
                if e.status == 404:
                    raise ExperimentNotFoundError(experiment_guid)

                raise UnknownKluAPIError(e.status, e.message)
            except Exception as e:
                raise UnknownKluError(e)



WORKSPACE_ENDPOINT = "/workspace/"
SINGLE_WORKSPACE_ENDPOINT = WORKSPACE_ENDPOINT + "{id}"
WORKSPACE_APPS_ENDPOINT = SINGLE_WORKSPACE_ENDPOINT + "/apps"
WORKSPACE_MODELS_ENDPOINT = SINGLE_WORKSPACE_ENDPOINT + "/models"
WORKSPACE_CONTEXTS_ENDPOINT = SINGLE_WORKSPACE_ENDPOINT + "/context"

DEFAULT_WORKSPACE_PAGE_SIZE = 50



import aiofiles
import aiohttp
from aiohttp import ClientResponse

from klu.context.models import PreSignUrlPostData


async def upload_to_pre_signed_url(
    session: aiohttp.ClientSession,
    pre_signed_url_data: PreSignUrlPostData,
    file_path: str,
) -> ClientResponse:
    pre_signed_url, pre_signed_fields = (
        pre_signed_url_data.url,
        pre_signed_url_data.fields,
    )

    data = aiohttp.FormData()
    for key, value in pre_signed_fields.items():
        data.add_field(key, value)

    async with aiofiles.open(file_path, mode="rb") as file:
        data.add_field("file", await file.read())

    async with session.post(pre_signed_url, data=data) as response:
        response.raise_for_status()
        return response



# mypy: disable-error-code="override"
from typing import NoReturn, Optional

import aiohttp
from aiohttp import ClientResponseError

from klu.common.client import KluClientBase
from klu.common.errors import (
    InvalidUpdateParamsError,
    NotSupportedError,
    UnknownKluAPIError,
    UnknownKluError,
)
from klu.finetune.constants import (
    FINE_TUNES_ENDPOINT,
    PROCESS_FINE_TUNE_ENDPOINT,
    READ_FINE_TUNE_STATUS_ENDPOINT,
)
from klu.finetune.errors import FineTuneNotFoundError
from klu.finetune.models import FineTune, FineTuneStatusResponse
from klu.utils.dict_helpers import dict_no_empty


class FineTunesClient(KluClientBase):
    def __init__(self, api_key: str):
        super().__init__(api_key, FINE_TUNES_ENDPOINT, FineTune)

    async def create(self) -> NoReturn:
        raise NotSupportedError()

    async def get(self, guid: str) -> FineTune:
        """
        Retrieves fine_tune information based on the unique FineTune guid, created during the FineTune creation.

        Args:
            guid (str): GUID of a fine_tune to fetch. The one that was used during the fine_tune creation

        Returns:
            FineTune object
        """
        return await super().get(guid)

    async def update(
        self,
        guid: str,
        name: Optional[str] = None,
        openai_fine_tune_name: Optional[str] = None,
    ) -> FineTune:
        """
        Update fine_tune data. At least one of the params has to be provided

        Args:
            guid (str): GUID of a fine_tune to update.
            name: Optional[str]. New fine_tune name
            openai_fine_tune_name: Optional[str]. New fine_tune openai_name

        Returns:
            Updated fine_tune instance
        """

        if not name and not openai_fine_tune_name:
            raise InvalidUpdateParamsError()

        return await super().update(
            **{
                "guid": guid,
                **dict_no_empty(
                    {
                        "name": name,
                        "openai_finetune_name": openai_fine_tune_name,
                    }
                ),
            }
        )

    async def delete(self, guid: str) -> FineTune:
        """
        Delete fine_tune based on the id.

        Args:
            guid (str): Unique Guid of a fine_tune to delete.

        Returns:
            Deleted fine_tune object
        """
        return await super().delete(guid)

    async def process(
        self,
        fine_tune_guid: str,
        base_model: str,
    ) -> FineTune:
        """
        Process the fine_tune.

        Args:
            fine_tune_guid (str): The GUID of the fine_tune to process.
            base_model (str): Can be one of [ada, babbage, curie, davinci] or a fine-tuned model created by your organization.

        Returns:
            The FineTune object
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)

            try:
                response = await client.post(
                    PROCESS_FINE_TUNE_ENDPOINT,
                    {
                        "baseModel": base_model,
                        "finetuneGuid": fine_tune_guid,
                    },
                )
                return FineTune._from_engine_format(response)
            except ClientResponseError as e:
                if e.status == 404:
                    raise FineTuneNotFoundError(fine_tune_guid)

                raise UnknownKluAPIError(e.status, e.message)
            except Exception as e:
                raise UnknownKluError(e)

    async def read_status(
        self,
        fine_tune_guid: str,
    ) -> FineTuneStatusResponse:
        """
        Read fine_tune status.

        Args:
            fine_tune_guid (str): The GUID of the fine_tune to process.

        Returns:
            The status of a FineTune and OpenAI FineTune name
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)

            try:
                response = await client.get(
                    READ_FINE_TUNE_STATUS_ENDPOINT.format(id=fine_tune_guid)
                )
                return FineTuneStatusResponse._create_instance(**response)
            except ClientResponseError as e:
                if e.status == 404:
                    raise FineTuneNotFoundError(fine_tune_guid)

                raise UnknownKluAPIError(e.status, e.message)
            except Exception as e:
                raise UnknownKluError(e)



from klu.common.errors import BaseKluError


class FeedbackNotFoundError(BaseKluError):
    feedback_id: int

    def __init__(self, feedback_id: int):
        self.feedback_id = feedback_id
        self.message = f"Feedback with id {feedback_id} was not found."
        super().__init__(self.message)



# mypy: disable-error-code="override"
from typing import Optional

from klu.common.client import KluClientBase
from klu.feedback.constants import FEEDBACK_ENDPOINT
from klu.feedback.models import Feedback
from klu.utils.dict_helpers import dict_no_empty


class FeedbackClient(KluClientBase):
    def __init__(self, api_key: str):
        super().__init__(api_key, FEEDBACK_ENDPOINT, Feedback)

    async def create(
        self,
        type: str,
        value: str,
        data_guid: int,
        created_by: str,
        metadata: Optional[dict] = None,
        source: str = 'api',
    ) -> Feedback:
        """
        Creates new feedback on a data point.

        Args:
            type (str): Type of feedback
            value (str): Value of feedback
            data_guid (int): ID of a data point to create feedback on
            created_by (str): User who created the feedback
            meta_data (dict): Metadata of feedback
            source (str): Source of feedback

        Returns:
            Created Feedback object
        """
        return await super().create(
            type=type,
            value=value,
            data=data_guid,
            createdById=created_by,
            source=source,
            meta_data=metadata,
        )

    async def get(self, guid: str) -> Feedback:
        """
        Retrieves data information based on the data ID.

        Args:
            guid (str): ID of a datum object to fetch.

        Returns:
            An object
        """
        return await super().get(guid)

    async def update(
        self,
        guid: str,
        type: Optional[str] = None,
        value: Optional[str] = None,
        source: Optional[str] = None,
        created_by: Optional[str] = None,
        metadata: Optional[dict] = None,
    ) -> Feedback:
        """
        Update feedback information.

        Args:
            guid (str): ID of a feedback object to update.
            type (str): Type of feedback
            value (str): Value of feedback
            source (str): Source of feedback
            created_by (str): User who created the feedback
            metadata (dict): Metadata of feedback

        Returns:
            Newly updated Feedback object
        """
        data = {
            "type": type,
            "value": value,
            "source": source,
            "created_by": created_by,
            "metadata": metadata,
        }
        return await super().update(**{"guid": guid, **dict_no_empty(data)})

    async def delete(self, guid: str) -> Feedback:
        return await super().delete(guid)



import json
from typing import AsyncIterator

from aiohttp_sse_client import client as sse_client

from klu.api.constants import (
    BEGIN_STREAM_CONTENT,
    END_STREAM_CONTENT,
    NO_MESSAGES_CONTENT,
)
from klu.common.errors import UnknownKluError


class SSEClient:
    streaming_url: str

    def __init__(self, streaming_url: str):
        self.streaming_url = streaming_url

    async def get_streaming_data(self) -> AsyncIterator[str]:
        """
        Get a streams of messages from an SQS queue for a specific channel.

        Returns:
            An asynchronous generator, which can be used to read chunks of data from the streaming url. Usage example:
                async for chunk in sse_client.get_streaming_data():
                    # Process the chunk of data here
        """
        try:
            async with sse_client.EventSource(self.streaming_url) as event_source:
                async for event in event_source:
                    message = event.data
                    if message == END_STREAM_CONTENT or message == NO_MESSAGES_CONTENT:
                        await event_source.close()
                        break

                    if message != BEGIN_STREAM_CONTENT:
                        token = json.loads(message)["token"]
                        yield token
        except Exception as e:
            raise UnknownKluError(e)



# mypy: disable-error-code="override"
from typing import List, NoReturn

import aiohttp

from klu.common.client import KluClientBase
from klu.common.errors import NotSupportedError
from klu.skill.constants import SKILLS_ENDPOINT
from klu.skill.models import Skill
from klu.utils.paginator import Paginator


class SkillsClient(KluClientBase):
    def __init__(self, api_key: str):
        super().__init__(
            api_key,
            SKILLS_ENDPOINT,
            Skill,
        )

    async def create(self) -> NoReturn:
        raise NotSupportedError()

    async def get(self, guid: str) -> Skill:
        """
        Retrieves skill info

        Args:
            guid (str): GUID of a skill to fetch

        Returns:
            Skill object
        """
        return await super().get(guid)

    async def update(self) -> NoReturn:
        raise NotSupportedError()

    async def delete(self, guid: str) -> Skill:
        """
        Delete skill based on the id.

        Args:
            guid (str): Unique Guid of a session to delete.

        Returns:
            Deleted session object
        """
        return await super().delete(guid)

    async def list(self) -> List[Skill]:
        """
        Retrieves all apps attached to a workspace.
        Does not rely on internal paginator state, so `reset_pagination` method call can be skipped

        Returns:
            An array of all apps
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            response = await self._paginator.fetch_all(client)

            return [Skill._from_engine_format(skill) for skill in response]

    async def reset_pagination(self):
        self._paginator = Paginator(SKILLS_ENDPOINT)



from typing import Optional


class BaseKluError(Exception):
    __suppress_context__ = True


class IdNotProvidedError(BaseKluError):
    def __init__(self):
        self.message = f"Received invalid parameters. id cannot be None. Please, provide a valid id."
        super().__init__(self.message)


class InstanceRelationshipNotFoundError(BaseKluError):
    def __init__(self, instance_name: str, message: str):
        self.message = f"One of the relationships for new {instance_name} was not found. Message from server: {message}"
        super().__init__(self.message)


class InstanceNotFoundError(BaseKluError):
    def __init__(self, instance_name: str, instance_id: str):
        self.message = f"{instance_name} with id {instance_id} was not found."
        super().__init__(self.message)

    def __str__(self):
        return f"{self.__class__.__name__}: {self.message}"


class UnknownKluError(BaseKluError):
    def __init__(self, error_message: Exception):
        self.message = f"Unknown error in Klu SDK. Please contact the support team."

        if error_message:
            self.message = f"{self.message}\nError message: {str(error_message)}"

        super().__init__(self.message)


class UnknownKluAPIError(BaseKluError):
    def __init__(self, status: int, message: str):
        self.message = (
            f"Unknown error in Klu API.\nstatus_code: {status},\nmessage: {message}"
        )
        super().__init__(self.message)


class UnauthorizedError(BaseKluError):
    def __init__(self):
        self.message = (
            f"Wrong credentials used to access the API. "
            f"Please, check you set the correct API key or contact the support team."
        )
        super().__init__(self.message)


class InvalidApiMethodUsedError(BaseKluError):
    def __init__(self, status: int, message: str):
        self.message = (
            "Invalid API method used. This is the result of internal programming error."
            f"Please, contact Klu support team so we could fix this.\nstatus_code: {status},\nmessage: {message}"
        )
        super().__init__(self.message)


class InvalidDataSent(BaseKluError):
    def __init__(self, status: int, message: str):
        self.message = (
            "Invalid model data was sent for creation or update. This is most likely the result of internal error."
            f"Please, contact Klu support team so we could fix this.\nstatus_code: {status},\nmessage: {message}"
        )
        super().__init__(self.message)


class BadRequestAPIError(BaseKluError):
    def __init__(self, status: int, message: str):
        self.message = (
            f"BadRequest error in Klu API.\nstatus_code: {status},\nmessage: {message}"
        )
        super().__init__(self.message)


class InvalidUpdateParamsError(BaseKluError):
    def __init__(self):
        self.message = "No update params have been provided. At least one of parameters should be sent."
        super().__init__(self.message)


class NotSupportedError(BaseKluError):
    def __init__(self):
        self.message = "This method is not supported"
        super().__init__(self.message)



from klu.common.errors import BaseKluError


class UnknownModelProviderError(BaseKluError):
    guid: str

    def __init__(self, guid):
        self.guid = guid
        self.message = (
            f"An unknown model or provider {self.guid} was used. "
            "Supported providers are [OpenAI, Azure OpenAI, Anthropic, GCP, NLPCloud, Cohere, AI21 & GooseAI]"
        )
        super().__init__(self.message)



"""
This module provides data models for the Session.
"""

from dataclasses import asdict, dataclass
from typing import Optional

from klu.common.models import BaseEngineModel


@dataclass
class Skill(BaseEngineModel):
    """
    This class represents the Session data returned from the Klu engine
    """

    guid: str
    name: str
    metadata: Optional[dict] = None
    created_at: Optional[str] = None
    updated_at: Optional[str] = None

    def __repr__(self):
        return self.generate_repr()

    @classmethod
    def _from_engine_format(cls, data: dict) -> "Skill":
        return cls._create_instance(
            **{
                "updated_at": data.pop("updatedAt", None),
                "created_at": data.pop("createdAt", None),
            },
            **data,
        )

    def _to_engine_format(self) -> dict:
        base_dict = asdict(
            self, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}
        )

        base_dict.pop("updated_at", None)
        base_dict.pop("created_at", None)

        return base_dict



CONTEXT_ENDPOINT = "/context/"
SINGLE_CONTEXT_ENDPOINT = CONTEXT_ENDPOINT + "{id}"
GENERATE_EMBEDDINGS = f"{CONTEXT_ENDPOINT}embed"
UPLOAD_PRE_SIGNED_URL_ENDPOINT = f"{CONTEXT_ENDPOINT}presign"
CONTEXT_STATUS_ENDPOINT = f"{SINGLE_CONTEXT_ENDPOINT}/status"
CONTEXT_SEARCH_ENDPOINT = f"{SINGLE_CONTEXT_ENDPOINT}/search"
CONTEXT_PROMPT_ENDPOINT = f"{CONTEXT_ENDPOINT}prompt"
CONTEXT_DOCUMENTS_ENDPOINT = f"{SINGLE_CONTEXT_ENDPOINT}/documents"
UPDATE_CONTEXT_DOCUMENT_ENDPOINT = f"{SINGLE_CONTEXT_ENDPOINT}/documents/" + "{doc_id}"



""" This module provides data models for the Workspace. """
from dataclasses import asdict, dataclass

from klu.common.models import BaseEngineModel


@dataclass
class Workspace(BaseEngineModel):
    """
    This class represents the Workspace data model returned from the Klu engine
    """

    id: str
    name: str
    slug: str
    created_by_id: str
    """
        Workspace unique identifier. This is the id you can later use to query this object and create app.
    """
    project_guid: str

    def __repr__(self):
        return self.generate_repr()

    @classmethod
    def _from_engine_format(cls, data: dict) -> "Workspace":
        return cls._create_instance(
            **{
                "created_by_id": data.pop("createdById", None),
            },
            **data,
        )

    def _to_engine_format(self) -> dict:
        base_dict = asdict(self)

        return {
            "createdById": base_dict.pop("created_by_id", None),
            **base_dict,
        }



ACTION_ENDPOINT = "/actions/"
CREATE_ACTION_ENDPOINT = f"{ACTION_ENDPOINT}create"
SINGLE_ACTION_ENDPOINT = ACTION_ENDPOINT + "{id}"
ACTION_DATA_ENDPOINT = f"{SINGLE_ACTION_ENDPOINT}/data"
PLAYGROUND_PROMPT_ENDPOINT = f"{ACTION_ENDPOINT}playground"
DEPLOY_ACTION_ENDPOINT = f"{SINGLE_ACTION_ENDPOINT}/deploy"
ACTION_SKILL_ENDPOINT = f"{SINGLE_ACTION_ENDPOINT}/skills"
ACTION_CONTEXT_ENDPOINT = f"{SINGLE_ACTION_ENDPOINT}/context"

DEFAULT_ACTIONS_PAGE_SIZE = 50



from klu.common.errors import BaseKluError


class AppNotFoundError(BaseKluError):
    app_id: str

    def __init__(self, app_id):
        self.app_id = app_id
        self.message = f"App with id {app_id} was not found."
        super().__init__(self.message)



from abc import ABCMeta, abstractmethod
from dataclasses import MISSING, dataclass, fields
from enum import Enum
from typing import Any, Dict, Optional, Union


class BasicEnum(Enum):
    """
    Basic class for enums across the projects. Adds basic functions to list names and values.
    Overrides comparison function to allow comparing against string values.
    """

    __hash__ = Enum.__hash__

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other.upper()
        return super().__eq__(other)

    @classmethod
    def list_values(cls) -> list:
        return list(map(lambda c: c.value, cls))

    @classmethod
    def list_names(cls) -> list:
        return list(map(lambda c: c.name, cls))


class BasicStringEnum(str, BasicEnum):
    """
    Basic class for string-based enums across the projects.
    Allows to have human-readable representation of enum as a value and convert between such value and key.
    """

    def __str__(self):
        return self.value

    @classmethod
    def get(cls, member: str) -> Optional["BasicStringEnum"]:
        """
        Get a member of an Enum by its string representation
        """
        member = member.upper().replace(" ", "_")
        try:
            return cls.__members__[member]
        except KeyError:
            return None

    def get_human_string(self):
        return self.value.replace("_", " ").capitalize() if self.value else None


@dataclass
class BaseDataClass:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    """
        This is the function that helps us to create an object in a way that any unknown kwargs passed will be ignored
    """

    @classmethod
    def _create_instance(cls, **kwargs):
        instance = cls.__new__(cls)
        instance._init_with_base_class(**kwargs)
        return instance

    def _init_with_base_class(self, **kwargs):
        for field_obj in fields(self):
            if field_obj.name in kwargs:
                value = kwargs[field_obj.name]
            elif field_obj.default is not MISSING:
                value = field_obj.default
            elif field_obj.default_factory is not MISSING:
                value = field_obj.default_factory()
            else:
                # TODO consider adding sentry log here to spot
                #  missing data from engine response (reflecting models discrepancy) in advance.
                # logging.warning(
                #     f"Missing required field {field_obj.name}. The field has been set to None by default. "
                #     "If you are unsure why this field is missing, please contact our team."
                # )
                value = None
            setattr(self, field_obj.name, value)

    def generate_repr(self):
        repr_str = f"{self.__class__.__name__}("
        for field in fields(self):
            field_value = getattr(self, field.name, "undefined")
            repr_str += f"{field.name}={field_value}, "
        repr_str = repr_str.rstrip(", ")
        repr_str += ")"
        return repr_str


class BaseEngineModel(BaseDataClass, metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def _from_engine_format(cls, data: dict) -> "BaseEngineModel":
        pass

    @abstractmethod
    def _to_engine_format(self) -> dict:
        pass


class TaskStatusEnum(BasicStringEnum):
    """The enum used to represent the status of any processing task status. Based on celery statuses."""

    # The task is waiting for execution.
    PENDING = "PENDING"
    # The task has been started.
    STARTED = "STARTED"

    # The task is to be retried, possibly because of failure.
    RETRY = "RETRY"
    # The task raised an exception, or has exceeded the retry limit.
    FAILURE = "FAILURE"
    # The task executed successfully.
    SUCCESS = "SUCCESS"


"""
This type defines input that can be sent to a prompt.
It can either be a string or a Dict with properties you defined in Klu
"""
PromptInput = Union[str, Dict[str, Any]]



from klu.action.client import ActionsClient
from klu.app.client import AppsClient
from klu.context.client import ContextClient
from klu.data.client import DataClient
from klu.experiment.client import ExperimentClient
from klu.feedback.client import FeedbackClient
from klu.finetune.client import FineTunesClient
from klu.model.client import ModelsClient
from klu.session.client import SessionClient
from klu.skill.client import SkillsClient
from klu.workspace.client import WorkspaceClient


class Klu:
    def __init__(self, api_key: str):
        self.data = DataClient(api_key)
        self.models = ModelsClient(api_key)
        self.actions = ActionsClient(api_key)
        self.context = ContextClient(api_key)
        self.sessions = SessionClient(api_key)
        self.workspace = WorkspaceClient(api_key)
        self.finetune = FineTunesClient(api_key)
        self.experiments = ExperimentClient(api_key)
        self.apps = AppsClient(api_key)
        self.feedback = FeedbackClient(api_key)
        self.skills = SkillsClient(api_key)



from klu.common.errors import BaseKluError


class ContextNotFoundError(BaseKluError):
    context_id: int

    def __init__(self, context_id):
        self.context_id = context_id
        self.message = f"Context with id {context_id} was not found."
        super().__init__(self.message)



class FineTuneNotFoundError(Exception):
    fine_tune_id: str

    def __init__(self, fine_tune_id: str):
        self.fine_tune_id = fine_tune_id
        self.message = f"FineTune with id {fine_tune_id} was not found."
        super().__init__(self.message)



# mypy: disable-error-code="override"
from typing import List, Optional

import aiohttp
from aiohttp import ClientResponseError

from klu.action.models import Action
from klu.app.constants import (
    APP_ACTIONS_ENDPOINT,
    APP_DATA_ENDPOINT,
    APP_ENDPOINT,
    DEFAULT_APPS_PAGE_SIZE,
)
from klu.app.errors import AppNotFoundError
from klu.app.models import App
from klu.common.client import KluClientBase
from klu.common.errors import InvalidUpdateParamsError, UnknownKluAPIError
from klu.data.models import Data
from klu.utils.dict_helpers import dict_no_empty
from klu.utils.paginator import Paginator


class AppsClient(KluClientBase):
    def __init__(self, api_key: str):
        super().__init__(api_key, APP_ENDPOINT, App)
        self._paginator = Paginator(APP_ENDPOINT)

    async def create(self, name: str, app_type: str, description: str) -> App:
        """
        Creates new app instance

        Args:
            name: str. Name of a new app
            app_type: str. Type of new app
            description: str. Description of a new app

        Returns:
            Newly created App object
        """
        return await super().create(
            name=name,
            app_type=app_type,
            description=description,
        )

    async def get(self, guid: str) -> App:
        """
        Retrieves app  information based on the app id.

        Args:
            guid (str): GUID of an app to fetch. The one that was used during the app creation

        Returns:
            App object
        """
        return await super().get(guid)

    async def update(
        self,
        guid: str,
        name: Optional[str] = None,
        app_type: Optional[str] = None,
        description: Optional[str] = None,
    ) -> App:
        """
        Update app metadata. At least one of the params has to be provided

        Args:
            guid (str): GUID of an app to fetch. The one that was used during the app creation
            name: Optional[str]. New app name
            app_type: Optional[str]. New app type
            description: Optional[str]. New app description

        Returns:
            Updated app instance
        """

        if not name and not app_type and not description:
            raise InvalidUpdateParamsError()

        return await super().update(
            **{
                "guid": guid,
                **dict_no_empty(
                    {"name": name, "app_type": app_type, "description": description}
                ),
            }
        )

    async def delete(self, guid: str) -> App:
        """
        Delete existing app information defined by the app id.

        Args:
            guid (str): The id of an app to delete.

        Returns:
            Deleted app object
        """
        return await super().delete(guid)

    async def get_data(self, app_guid: str) -> List[Data]:
        """
        Retrieves app data information based on the app GUID.

        Args:
            app_guid (str): GUID of an app to fetch data for. The one that was returned during the app creation

        Returns:
            An array of data objects, found by provided app id.
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.get(APP_DATA_ENDPOINT.format(id=app_guid))
                return [Data._from_engine_format(data) for data in response]
            except ClientResponseError as e:
                if e.status == 404:
                    raise AppNotFoundError(app_guid)

                raise UnknownKluAPIError(e.status, e.message)

    async def get_actions(self, app_guid: str) -> List[Action]:
        """
        Retrieves app actions information based on the app GUID.

        Args:
            app_guid (str): GUID of an app to fetch actions for. The one that was used during the app creation

        Returns:
            An array of actions, found by provided app id.
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.get(APP_ACTIONS_ENDPOINT.format(id=app_guid))
                return [Action._from_engine_format(action) for action in response]
            except ClientResponseError as e:
                if e.status == 404:
                    raise AppNotFoundError(app_guid)

                raise UnknownKluAPIError(e.status, e.message)

    async def list(self) -> List[App]:
        """
        Retrieves all apps attached to a workspace.
        Does not rely on internal paginator state, so `reset_pagination` method call can be skipped

        Returns:
            An array of all apps
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            response = await self._paginator.fetch_all(client)

            return [App._from_engine_format(app) for app in response]

    async def fetch_single_page(
        self, page_number, limit: int = DEFAULT_APPS_PAGE_SIZE
    ) -> List[App]:
        """
        Retrieves a single page of apps.
        Can be used to fetch a specific page of apps provided a certain per_page config.
        Does not rely on internal paginator state, so `reset_pagination` method call can be skipped

        Args:
            page_number (int): Number of the page to fetch
            limit (int): Number of instances to fetch per page. Defaults to 50

        Returns:
            An array of apps fetched for a queried page. Empty if queried page does not exist
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            response = await self._paginator.fetch_single_page(
                client, page_number, limit=limit
            )

            return [App._from_engine_format(app) for app in response]

    async def fetch_next_page(
        self, limit: int = DEFAULT_APPS_PAGE_SIZE, offset: Optional[int] = None
    ) -> List[App]:
        """
        Retrieves the next page of apps. Can be used to fetch a flexible number of pages starting.
        The place to start from can be controlled by the offset parameter.
        After using this method, we suggest to call `reset_pagination` method to reset the page cursor.

        Args:
            limit (int): Number of instances to fetch per page. Defaults to 50
            offset (int): The number of apps to skip. Can be used to query the pages of apps skipping certain number of instances.

        Returns:
            An array of apps fetched for a queried page. Empty if the end was reached at the previous step.
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            response = await self._paginator.fetch_next_page(
                client, limit=limit, offset=offset
            )

            return [App._from_engine_format(app) for app in response]

    async def reset_pagination(self):
        self._paginator = Paginator(APP_ENDPOINT)



"""
This module provides data models for the Data.
"""

from abc import ABC
from dataclasses import asdict, dataclass
from typing import Any, Optional

from klu.common.models import BaseEngineModel, BasicStringEnum


@dataclass
class DataWithId(BaseEngineModel, ABC):
    """
    This class represents the base data with Id after it has been persisted into the Klu Database
    """

    guid: str

    def __repr__(self):
        return self.generate_repr()


@dataclass
class DataWithFeedbackUrl(BaseEngineModel, ABC):
    """
    This class represents the data specific to the model returned from the Action prompting
    """

    feedback_url: int

    def __repr__(self):
        return self.generate_repr()


@dataclass
class DataBaseClass(BaseEngineModel):
    """
    This class represents the generic data information that is stored in the Klu database
    """

    action: str
    input: str
    output: str
    session: Optional[str] = None
    metadata: Optional[dict] = None
    model: Optional[str] = None
    model_provider: Optional[str] = None
    full_prompt_sent: Optional[dict] = None
    system_message: Optional[str] = None
    latency: Optional[int] = None
    num_output_tokens: Optional[int] = None
    num_input_tokens: Optional[int] = None
    feedback: Optional[Any] = None

    def __repr__(self):
        return self.generate_repr()

    @classmethod
    def _from_engine_format(cls, data: dict) -> "DataBaseClass":
        return cls._create_instance(
            **{
                "updated_at": data.pop("updatedAt", None),
                "created_at": data.pop("createdAt", None),
            },
            **data,
        )

    def _to_engine_format(self) -> dict:
        base_dict = asdict(
            self, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}
        )

        base_dict.pop("updated_at", None)
        base_dict.pop("created_at", None)

        return base_dict


# We need this way of inheritance to append optional field after the required ones.
@dataclass
class Data(DataBaseClass, DataWithId):
    """
    This class represents the Data model returned from the Klu engine
    """

    @classmethod
    def _from_engine_format(cls, data: dict) -> "Data":
        return cls._create_instance(
            **{
                "updated_at": data.pop("updatedAt", None),
                "created_at": data.pop("createdAt", None),
            },
            **data,
        )

    def _to_engine_format(self) -> dict:
        base_dict = asdict(
            self, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}
        )

        base_dict.pop("updated_at", None)
        base_dict.pop("created_at", None)

        return base_dict


@dataclass
class ActionData(DataBaseClass, DataWithFeedbackUrl):
    """
    This class represents the Data model returned from the Klu engine in response to Action data request
    """

    @classmethod
    def _from_engine_format(cls, data: dict) -> "ActionData":
        return cls._create_instance(
            **{
                "updated_at": data.pop("updatedAt", None),
                "created_at": data.pop("createdAt", None),
            },
            **data,
        )

    def _to_engine_format(self) -> dict:
        base_dict = asdict(
            self, dict_factory=lambda x: {k: v for (k, v) in x if v is not None}
        )

        base_dict.pop("updated_at", None)
        base_dict.pop("created_at", None)

        return base_dict


class DataSourceType(BasicStringEnum):
    """The enum used to represent the source of the Data"""

    SDK = "SDK"
    API = "API"
    BACK_FILL = "backfill"



class ActionNotFoundError(Exception):
    action_id: str

    def __init__(self, action_id: str):
        self.action_id = action_id
        self.message = f"Action with id {action_id} was not found."
        super().__init__(self.message)


class InvalidActionPromptData(Exception):
    def __init__(self, response_message: str):
        self.message = (
            f"Failed to run the action due to the invalid request parameters. "
            f"Response message: {response_message}"
        )
        super().__init__(self.message)



# mypy: disable-error-code="override"
from typing import List, Optional

import aiohttp
from aiohttp import ClientResponseError

from klu.app.models import App
from klu.common.client import KluClientBase
from klu.common.errors import UnknownKluAPIError
from klu.context.models import Context
from klu.model.models import Model
from klu.utils.paginator import Paginator
from klu.workspace.constants import (
    DEFAULT_WORKSPACE_PAGE_SIZE,
    WORKSPACE_APPS_ENDPOINT,
    WORKSPACE_CONTEXTS_ENDPOINT,
    WORKSPACE_ENDPOINT,
    WORKSPACE_MODELS_ENDPOINT,
)
from klu.workspace.errors import WorkspaceOrUserNotFoundError
from klu.workspace.models import Workspace


class WorkspaceClient(KluClientBase):
    def __init__(self, api_key: str):
        super().__init__(api_key, WORKSPACE_ENDPOINT, Workspace)
        self._paginator = Paginator(WORKSPACE_ENDPOINT)

    async def create(self, name: str, slug: str) -> Workspace:
        """
        Creates a Workspace based on the data provided.

        Args:
            name (str): Model key. Required
            slug (str): Workspace slug. The unique name you would prefer to use to access the model.

        Returns:
            A newly created Workspace object.
        """
        return await super().create(name=name, slug=slug)

    async def get(self, guid: str) -> Workspace:
        """
        Retrieves a single workspace object by provided workspace id

        Args:
            guid (str): The ID of a workspace to fetch. project_guid you sent during the workspace creation

        Returns:
            A workspace object
        """
        return await super().get(guid)

    async def update(self, guid: str, name: str) -> Workspace:
        """
        Update workspace data. Currently, only name update is supported.

        Args:
            guid (str): ID of a context to update. project_guid you sent during the workspace creation
            name: str. New workspace name.

        Returns:
            Updated workspace instance
        """
        return await super().update(guid=guid, name=name)

    async def delete(self, guid: str) -> Workspace:
        """
        Delete Workspace based on the id.

        Args:
            guid (str): ID of a workspace to delete. project_guid you sent during the workspace creation

        Returns:
            Deleted workspace object
        """
        return await super().delete(guid)

    async def get_apps(self, guid: str) -> List[App]:
        """
        Retrieves the list of apps for workspace defined by provided guid

        Args:
            guid (str): The ID of workspace to fetch apps for.

        Returns:
            List of apps found in a workspace
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.get(
                    WORKSPACE_APPS_ENDPOINT.format(id=guid),
                )
                return [App._from_engine_format(app) for app in response]
            except ClientResponseError as e:
                if e.status == 404:
                    raise WorkspaceOrUserNotFoundError()

                raise UnknownKluAPIError(e.status, e.message)

    async def get_contexts(self, guid: str) -> List[Context]:
        """
        Retrieves the list of contexts for workspace defined by provided guid

        Args:
            guid (str): The ID of workspace to fetch contexts for.
                project_guid you sent during the workspace creation

        Returns:
            List of Context objects found on a workspace.
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.get(
                    WORKSPACE_CONTEXTS_ENDPOINT.format(id=guid),
                )
                return [Context._from_engine_format(context) for context in response]
            except ClientResponseError as e:
                if e.status == 404:
                    raise WorkspaceOrUserNotFoundError()

                raise UnknownKluAPIError(e.status, e.message)

    async def get_models(self, guid: str) -> List[Model]:
        """
        Retrieves the list of models for provided workspace id

        Args:
            guid (str): The ID of workspace to fetch models for. project_guid you sent during the workspace creation

        Returns:
            List of Model objects found on a workspace.
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.get(
                    WORKSPACE_MODELS_ENDPOINT.format(id=guid),
                )
                return [Model._from_engine_format(model) for model in response]
            except ClientResponseError as e:
                if e.status == 404:
                    raise WorkspaceOrUserNotFoundError()

                raise UnknownKluAPIError(e.status, e.message)

    async def list(self) -> List[Workspace]:
        """
        Retrieves all workspaces for a user represented by the used API_KEY.
        Does not rely on internal paginator state, so `reset_pagination` method call can be skipped

        Returns:
            An array of all workspaces
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            response = await self._paginator.fetch_all(client)

            return [Workspace._from_engine_format(workspace) for workspace in response]

    async def fetch_single_page(
        self, page_number, limit: int = DEFAULT_WORKSPACE_PAGE_SIZE
    ) -> List[Workspace]:
        """
        Retrieves a single page of workspaces.
        Can be used to fetch a specific page of workspaces provided a certain per_page config.
        Does not rely on internal paginator state, so `reset_pagination` method call can be skipped

        Args:
            page_number (int): Number of the page to fetch
            limit (int): Number of instances to fetch per page. Defaults to 50

        Returns:
            An array of workspaces fetched for a queried page. Empty if queried page does not exist
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            response = await self._paginator.fetch_single_page(
                client, page_number, limit=limit
            )

            return [Workspace._from_engine_format(workspace) for workspace in response]

    async def fetch_next_page(
        self, limit: int = DEFAULT_WORKSPACE_PAGE_SIZE, offset: Optional[int] = None
    ) -> List[Workspace]:
        """
        Retrieves the next page of workspaces. Can be used to fetch a flexible number of pages starting.
        The place to start from can be controlled by the offset parameter.
        After using this method, we suggest to call `reset_pagination` method to reset the page cursor.

        Args:
            limit (int): Number of instances to fetch per page. Defaults to 50
            offset (int): The number of instances to skip. Can be used to query the pages of workspaces skipping certain number of instances.

        Returns:
            An array of workspaces fetched for a queried page. Empty if the end was reached at the previous step.
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            response = await self._paginator.fetch_next_page(
                client, limit=limit, offset=offset
            )

            return [Workspace._from_engine_format(workspace) for workspace in response]

    async def reset_pagination(self):
        self._paginator = Paginator(WORKSPACE_ENDPOINT)



from functools import wraps
from typing import List, Optional, Union

import aiohttp
from aiohttp import ClientResponseError

from klu.api.config import get_api_url
from klu.common.errors import (
    BadRequestAPIError,
    InvalidApiMethodUsedError,
    InvalidDataSent,
    UnauthorizedError,
    UnknownKluAPIError,
    UnknownKluError,
)


def _handle_http_exception(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except ClientResponseError as e:
            if e.status == 400:
                raise BadRequestAPIError(e.status, e.message)
            if e.status == 500:
                raise UnknownKluAPIError(e.status, e.message)
            if e.status == 401:
                raise UnauthorizedError()
            if e.status == 405:
                raise InvalidApiMethodUsedError(e.status, e.message)
            if e.status == 422:
                raise InvalidDataSent(e.status, e.message)

            # Passing higher for more specific handling up to the functions level.
            raise e
        except Exception as e:
            raise UnknownKluError(e)

    return wrapper


class APIClient:
    def __init__(self, session: aiohttp.ClientSession, api_key: str):
        self.session = session
        self.api_url = get_api_url()
        self.headers = {"Authorization": f"Bearer {api_key}"}

    @_handle_http_exception
    async def get(
        self, path: str, params: Optional[dict] = None, api_url: Optional[str] = None
    ) -> Union[dict, List[dict]]:
        url = f"{self.api_url if api_url is None else api_url}{path}"
        async with self.session.get(
            url, params=params, headers=self.headers
        ) as response:
            response.raise_for_status()
            return await response.json()

    @_handle_http_exception
    async def post(self, path: str, json_data: Optional[dict] = None) -> dict:
        url = f"{self.api_url}{path}"
        async with self.session.post(
            url, json=json_data, headers=self.headers
        ) as response:
            if not response.ok and response.content_type == "application/json":
                content = await response.json()
                detail_message = content.get("detail") if content else None

                if detail_message:
                    response.reason = detail_message

            response.raise_for_status()
            return await response.json()

    @_handle_http_exception
    async def put(self, path: str, json_data: dict) -> dict:
        url = f"{self.api_url}{path}"
        async with self.session.put(
            url, json=json_data, headers=self.headers
        ) as response:
            response.raise_for_status()
            return await response.json()

    @_handle_http_exception
    async def delete(self, path: str) -> dict:
        url = f"{self.api_url}{path}"
        async with self.session.delete(url, headers=self.headers) as response:
            response.raise_for_status()
            return await response.json()



"""
This module provides data models for the Context.
"""
from dataclasses import asdict, dataclass
from typing import Any, List, Optional, Union

from klu.common.models import BaseDataClass, BaseEngineModel


@dataclass
class Context(BaseEngineModel):
    """
    This class represents the Context model returned from the Klu engine
    """

    guid: str
    name: str
    processed: bool

    type_id: int
    task_id: str
    index_url: str
    loader_id: int
    workspace_id: int
    created_by_id: str

    description: Optional[str]
    updated_at: Optional[str] = None
    created_at: Optional[str] = None
    meta_data: Optional[Union[dict, list]] = None

    def __repr__(self):
        return self.generate_repr()

    @classmethod
    def _from_engine_format(cls, data: dict) -> "Context":
        return cls._create_instance(
            **{
                "type_id": data.pop("typeId", None),
                "task_id": data.pop("taskId", None),
                "files": data.pop("files", None),
                "index_url": data.pop("indexUrl", None),
                "loader_id": data.pop("loaderId", None),
                "updated_at": data.pop("updatedAt", None),
                "created_at": data.pop("createdAt", None),
                "workspace_id": data.pop("workspaceId", None),
                "created_by_id": data.pop("createdById", None),
            },
            **data,
        )

    def _to_engine_format(self) -> dict:
        base_dict = asdict(self)

        return {
            "typeId": base_dict.pop("type_id", None),
            "taskId": base_dict.pop("task_id", None),
            "files": base_dict.pop("files", None),
            "indexUrl": base_dict.pop("index_url", None),
            "loaderId": base_dict.pop("loader_id", None),
            "updatedAt": base_dict.pop("updated_at", None),
            "createdAt": base_dict.pop("created_at", None),
            "workspaceId": base_dict.pop("workspace_id", None),
            "createdById": base_dict.pop("created_by_id", None),
            **base_dict,
        }


@dataclass
class ContextDocument(BaseEngineModel):
    guid: str
    content: str
    created_at: Optional[str]
    updated_at: Optional[str]
    filter: Optional[str]
    meta_data: Optional[Any]
    score: Optional[float]

    def __repr__(self):
        return self.generate_repr()

    @classmethod
    def _from_engine_format(cls, data: dict) -> "ContextDocument":
        return cls._create_instance(
            **{
                "updated_at": data.pop("updatedAt", None),
                "created_at": data.pop("createdAt", None),
            },
            **data,
        )

    def _to_engine_format(self) -> dict:
        base_dict = asdict(self)

        return {
            **base_dict,
        }


@dataclass
class AddContextDocumentResponse(BaseDataClass):
    docs: List[str]
    status: str


@dataclass
class DeleteContextDocumentResponse(BaseDataClass):
    status: str


@dataclass
class ContextPromptResponse(BaseDataClass):
    """
    This class represents the Response data model returned from the Klu engine in response to action prompting.
    """

    msg: str
    streaming: bool
    data_guid: Optional[str] = None
    result_url: Optional[str] = None
    feedback_url: Optional[str] = None
    streaming_url: Optional[str] = None
    metadata: Optional[dict] = None

    def __repr__(self):
        return self.generate_repr()


@dataclass
class PreSignUrlPostData(BaseDataClass):
    """
    Data that represents response from pre-signed url generation.
    url - pre-signed url that can be used to upload the file.
    fields - dict with data that has to be passed alongside the file during the upload
    object_url - contains the url that can be used to access the file location after the upload.
    This same object_url can be used during the context creation.
    """

    url: str
    fields: dict
    object_url: str

    def __repr__(self):
        return self.generate_repr()


@dataclass
class FileData(BaseDataClass):
    """
    file_name (str): The name of the file to be uploaded. Has to be unique among the files you uploaded before.
    file_path (str): The path to a file in your system.
    The file path can be an absolute path, a relative path, or a path that includes variables like ~ or $HOME
    """

    file_path: str
    file_name: str

    def __repr__(self):
        return self.generate_repr()



from abc import abstractmethod
from typing import Type, TypeVar

import aiohttp
from aiohttp import ClientResponseError

from klu.api.client import APIClient
from klu.common.errors import (
    IdNotProvidedError,
    InstanceNotFoundError,
    InstanceRelationshipNotFoundError,
)
from klu.common.models import BaseEngineModel

T = TypeVar("T", bound=BaseEngineModel)


class KluClientBase:
    def __init__(self, api_key: str, base_path: str, model: Type[T]):
        self._model = model
        self._api_key = api_key
        self._base_path = base_path

    def _get_api_client(self, session):
        return APIClient(session, self._api_key)

    @abstractmethod
    async def create(self, **kwargs) -> T:
        # Some endpoints use a different url for creation
        # TODO ideally, this should be the same. Create a task on the engine side.
        url = kwargs.pop("url", None)
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.post(url or self._base_path, kwargs)
                return self._model._from_engine_format(response)  # type: ignore
            except ClientResponseError as e:
                if e.status == 404:
                    raise InstanceRelationshipNotFoundError(
                        self._model.__name__, e.message
                    )

                raise e

    @abstractmethod
    async def get(self, guid: str) -> T:
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.get(f"{self._base_path}{guid}")
                return self._model._from_engine_format(response)  # type: ignore
            except ClientResponseError as e:
                if e.status == 404:
                    raise InstanceNotFoundError(self._model.__name__, guid)

                raise e

    @abstractmethod
    async def update(self, **kwargs) -> T:
        id = kwargs.pop("guid", None)
        if not id:
            raise IdNotProvidedError()

        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.put(f"{self._base_path}{id}", kwargs)
                return self._model._from_engine_format(response)  # type: ignore
            except ClientResponseError as e:
                if e.status == 404:
                    raise InstanceNotFoundError(self._model.__name__, id)

                raise e

    @abstractmethod
    async def delete(self, guid: str) -> T:
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.delete(f"{self._base_path}{guid}")
                return self._model._from_engine_format(response)  # type: ignore
            except ClientResponseError as e:
                if e.status == 404:
                    raise InstanceNotFoundError(self._model.__name__, guid)

                raise e



from klu.common.errors import BaseKluError


class WorkspaceOrUserNotFoundError(BaseKluError):
    def __init__(self):
        self.message = f"User was not found Workspace for current user was not found"
        super().__init__(self.message)



"""
This module provides data models for the App.
"""
from dataclasses import asdict, dataclass
from datetime import datetime
from typing import Optional

from klu.common.models import BaseEngineModel


@dataclass
class App(BaseEngineModel):
    """
    This class represents the App data model returned from the Klu engine
    """

    id: int
    guid: str
    name: str
    app_type: str
    enabled: bool

    workspace_id: int
    created_by_id: str

    description: Optional[str]
    deleted: Optional[bool] = None
    updated_at: Optional[str] = None
    created_at: Optional[str] = None

    def __repr__(self):
        return self.generate_repr()

    @classmethod
    def _from_engine_format(cls, data: dict) -> "App":
        return cls._create_instance(
            **{
                "updated_at": data.pop("updatedAt", None),
                "created_at": data.pop("createdAt", None),
                "workspace_id": data.pop("workspaceId", None),
                "created_by_id": data.pop("createdById", None),
            },
            **data,
        )

    def _to_engine_format(self) -> dict:
        base_dict = asdict(self)

        return {
            "updatedAt": base_dict.pop("updated_at", None),
            "createdAt": base_dict.pop("created_at", None),
            "workspaceId": base_dict.pop("workspace_id", None),
            "createdById": base_dict.pop("created_by_id", None),
            **base_dict,
        }



# mypy: disable-error-code="override"
from typing import List, Optional

import aiohttp
from aiohttp import ClientResponseError

from klu.common.client import KluClientBase
from klu.common.errors import InvalidUpdateParamsError, UnknownKluAPIError
from klu.model.constants import (
    MODEL_DEFAULT_ENDPOINT,
    MODEL_ENDPOINT,
    PROVIDERS_ENDPOINT,
    VALIDATE_MODEL_API_KEY_ENDPOINT,
)
from klu.model.errors import UnknownModelProviderError
from klu.model.models import Model, Provider, ProviderWithModels
from klu.utils.dict_helpers import dict_no_empty


class ModelsClient(KluClientBase):
    def __init__(self, api_key: str):
        super().__init__(api_key, MODEL_ENDPOINT, Model)

    async def create(
        self,
        llm: str,
        provider: str,
        key: str,
        nickname: Optional[str] = "",
        url: Optional[str] = "",
        default: Optional[bool] = False,
    ) -> Model:
        """
                Creates a model based on the data provided.

                Args:
                    llm (str): Model llm. Required
                    workspace_model_provider_id (str): Unique identifier of model provider.
                        Can be retrieved from a model providers listing endpoint.

                Returns:
        A newly created Model object.
        """
        return await super().create(
            llm=llm,
            provider=provider,
            key=key,
            url=url,
            nickname=nickname,
            default=default,
        )

    async def get(self, guid: str) -> Model:
        """
        Retrieves model  information based on the id.

        Args:
            guid (str): GUID of a model to fetch. The one that was used during the model creation

        Returns:
            Model object
        """
        return await super().get(guid)

    async def update(
        self, guid: str, llm: Optional[str] = None, key: Optional[str] = None
    ) -> Model:
        """
        Update model data. At least one of the params has to be provided

        Args:
            guid (str): GUID of a model to update
            llm: Optional[str]. New model
            key: Optional[str]. New model key

        Returns:
            Updated app instance
        """
        if not llm and not key:
            raise InvalidUpdateParamsError()

        return await super().update(
            **{"guid": guid, **dict_no_empty({"llm": llm, "key": key})}
        )

    async def delete(self, guid: str) -> Model:
        """
        Delete model based on the id.

        Args:
            guid (str): ID of a model to delete.

        Returns:
            Deleted model object
        """
        return await super().delete(guid)

    async def validate_key(self, api_key: str, provider: str) -> bool:
        """
        Validates API keys for provided api_key and provider values.

        Args:
            api_key (str): Model api_key
            provider (int): Model provider id. Should be taken from the UI.

        Returns:
            A JSON response with a message about successful creation if model was created.
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.post(
                    VALIDATE_MODEL_API_KEY_ENDPOINT,
                    {
                        "api_key": api_key,
                        "provider": provider,
                    },
                )
                return bool(response.get("validated"))
            except ClientResponseError as e:
                if e.status == 404:
                    raise UnknownModelProviderError(provider)

                raise UnknownKluAPIError(e.status, e.message)

    async def add_provider(
        self,
        provider: str,
        key: str,
        url: Optional[str] = None,
        default: Optional[bool] = False,
    ) -> ProviderWithModels:
        """
                Creates a model based on the data provided.

                Args:
                    llm (str): Model llm. Required
                    workspace_model_provider_id (str): Unique identifier of model provider.
                        Can be retrieved from a model providers listing endpoint.

                Returns:
        A newly created Model object.
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.post(
                    PROVIDERS_ENDPOINT,
                    {
                        "provider": provider,
                        "key": key,
                        "url": url,
                        "default": default,
                    },
                )
                provider = Provider._from_engine_format(response.pop("provider"))  # type: ignore
                models = []
                for model in response["models"]:
                    temp = Model._from_engine_format(model)  # type: ignore
                    models.append(temp)  # type: ignore
                return ProviderWithModels(provider=provider, models=models)  # type: ignore
            except ClientResponseError as e:
                raise UnknownKluAPIError(e.status, e.message)

    async def update_provider(
        self,
        provider: str,
        nickname: Optional[str] = "",
        key: Optional[str] = None,
        url: Optional[str] = None,
        default: Optional[bool] = False,
    ) -> Provider:
        """
                Creates a model based on the data provided.

                Args:
                    llm (str): Model llm. Required
                    workspace_model_provider_id (str): Unique identifier of model provider.
                        Can be retrieved from a model providers listing endpoint.

                Returns:
        A newly created Model object.
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.put(
                    PROVIDERS_ENDPOINT + provider,
                    dict_no_empty(
                        {
                            "nickname": nickname,
                            "key": key,
                            "url": url,
                            "default": default,
                        }
                    ),
                )
                provider = Provider._from_engine_format(response.pop("provider"))  # type: ignore
                models = []
                for model in response["models"]:
                    temp = Model._from_engine_format(model)  # type: ignore
                    models.append(temp)  # type: ignore
                return ProviderWithModels(provider=provider, models=models)  # type: ignore
            except ClientResponseError as e:
                raise UnknownKluAPIError(e.status, e.message)

    async def get_providers(self) -> List[Provider]:
        """
        Retrievel all providers in the workspace

        Returns:
            List of Provider objects
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.get(
                    PROVIDERS_ENDPOINT,
                )
                providers = []
                for provider in response:
                    providers.append(Provider._from_engine_format(provider))  # type: ignore
                return providers  # type: ignore
            except ClientResponseError as e:
                raise UnknownKluAPIError(e.status, e.message)

    async def delete_provider(self, guid: str) -> bool:
        """
        Retrievel all providers in the workspace

        Returns:
            List of Model objects
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.delete(
                    PROVIDERS_ENDPOINT + guid,
                )
                if response == "Success":
                    return True
                else:
                    return False
            except ClientResponseError as e:
                raise UnknownKluAPIError(e.status, e.message)

    async def get_all(
        self, skip: Optional[int] = 0, limit: Optional[int] = 500
    ) -> List[Model]:
        """
        Retrieves all models in the workspace.
        By default will return the first 500.

        Params:
            skip (int): Number of models to skip
            limit (int): Number of models to return

        Returns:
            List of Model objects
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.get(
                    MODEL_ENDPOINT, {"skip": skip, "limit": limit}
                )
                models = []
                for model in response:
                    models.append(Model._from_engine_format(model))
                return models
            except ClientResponseError as e:
                raise UnknownKluAPIError(e.status, e.message)

    async def get_default(self) -> Model:
        """
        Retrievel the default model for the workspace

        Returns:
            Default model.
        """
        async with aiohttp.ClientSession() as session:
            client = self._get_api_client(session)
            try:
                response = await client.get(
                    MODEL_DEFAULT_ENDPOINT,
                )
                return Model._from_engine_format(response)
            except ClientResponseError as e:
                raise UnknownKluAPIError(e.status, e.message)



from typing import Optional

from klu.api.client import APIClient

DEFAULT_PAGE_SIZE = 100


class Paginator:
    def __init__(self, endpoint, limit=DEFAULT_PAGE_SIZE):
        self.endpoint = endpoint

        self.limit = limit
        self.current_page = 0
        self.total_count = None
        self.has_next_page = True

    async def _get_paginated_results(
        self,
        api_client: APIClient,
        skip: Optional[int] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        **params,
    ):
        limit = limit if limit is not None else self.limit
        skip = skip if skip is not None else self.current_page * limit

        if offset is not None:
            skip = skip + offset

        # Make API request with skip, limit, and additional parameters
        response = await api_client.get(
            self.endpoint, {**params, "skip": skip, "limit": limit}
        )
        results = response.get("data", [])

        self.total_count = response.get("total_count", 0)
        self.has_next_page = response.get("has_next_page", False)

        return results

    async def fetch_all(self, api_client: APIClient, limit: Optional[int] = None):
        results = []
        self._reset_paginator()

        while self.has_next_page:
            results.extend(await self._get_paginated_results(api_client, limit=limit))
            self.current_page += 1

        self._reset_paginator()
        return results

    async def fetch_single_page(
        self, api_client: APIClient, page_number: int, limit: Optional[int] = None
    ):
        self._reset_paginator()

        limit = self.limit if limit is None else limit
        skip = (page_number - 1) * limit

        results = await self._get_paginated_results(api_client, skip=skip, limit=limit)

        self._reset_paginator()
        return results

    async def fetch_next_page(
        self,
        api_client: APIClient,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ):
        if not self.has_next_page:
            return []

        curr_page_results = await self._get_paginated_results(
            api_client, limit=limit, offset=offset
        )
        self.current_page += 1

        return curr_page_results

    def _reset_paginator(self):
        self.current_page = 0
        self.total_count = None
        self.has_next_page = True



APP_ENDPOINT = "/apps/"
SINGLE_APP_ENDPOINT = APP_ENDPOINT + "{id}"
APP_DATA_ENDPOINT = f"{SINGLE_APP_ENDPOINT}/data"
APP_ACTIONS_ENDPOINT = f"{SINGLE_APP_ENDPOINT}/actions"

DEFAULT_APPS_PAGE_SIZE = 50
